#include <algorithm>
#include <cmath>
#include <msp/gl/technique.h>
#include "libr2c2/trackappearance.h"
#include "catalogue.h"
#include "tracktype.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

TrackType3D::TrackType3D(Catalogue3D &c, const TrackType &tt):
	ObjectType3D(c),
	mesh(0),
	object(0),
	own_data(false)
{
	const TrackAppearance &appearance = tt.get_appearance();
	const vector<TrackPart> &parts = tt.get_parts();

	const Profile &ballast_profile = appearance.get_ballast_profile();
	const Vector &ballast_min = ballast_profile.get_min_coords();
	const Vector &ballast_max = ballast_profile.get_max_coords();
	float ballast_h = ballast_max.y-ballast_min.y;

	const Profile &rail_profile = appearance.get_rail_profile();
	const Vector &rail_min = rail_profile.get_min_coords();
	const Vector &rail_max = rail_profile.get_max_coords();
	float rail_h = rail_max.y-rail_min.y;

	const Profile &tie_profile = appearance.get_tie_profile();
	const Vector &tie_min = tie_profile.get_min_coords();
	const Vector &tie_max = tie_profile.get_max_coords();
	float tie_h = tie_max.y-tie_min.y;

	float gauge = appearance.get_gauge();

	string obj_name = tt.get_object();
	if(!obj_name.empty())
		object = &catalogue.get<GL::Object>(obj_name);
	else
	{
		mesh = new GL::Mesh((GL::NORMAL3, GL::TEXCOORD2, GL::VERTEX3));
		mesh->set_winding(&GL::WindingTest::counterclockwise());
		GL::MeshBuilder bld(*mesh);

		unsigned index = 0;
		bld.texcoord(0.17, 0.5);
		for(vector<TrackPart>::const_iterator i=parts.begin(); i!=parts.end(); ++i)
			build_part(*i, ballast_profile, Vector(0, 0, -ballast_min.y), false, bld, index);

		float y = ballast_h-tie_min.y;
		bld.texcoord(0.83, 0.5);
		for(vector<TrackPart>::const_iterator i=parts.begin(); i!=parts.end(); ++i)
			build_ties(*i, tie_profile, appearance.get_tie_length(), appearance.get_tie_spacing(), Vector(0, 0, y), bld, index);

		bld.texcoord(0.5, 0.5);
		y = ballast_h+tie_h-rail_min.y;
		for(vector<TrackPart>::const_iterator i=parts.begin(); i!=parts.end(); ++i)
			build_part(*i, rail_profile, Vector(0, gauge/2, y), true, bld, index);
		for(vector<TrackPart>::const_iterator i=parts.begin(); i!=parts.end(); ++i)
			build_part(*i, rail_profile, Vector(0, -gauge/2, y), false, bld, index);

		object = new GL::Object;
		object->set_mesh(mesh);
		object->set_technique(&catalogue.get<GL::Technique>(appearance.get_technique()));

		own_data = true;
	}

	unsigned paths = tt.get_paths();
	for(int i=-1; i<=1; ++i)
	{
		// TODO Make profile width configurable
		Profile profile;
		if(i==0)
		{
			float rail_w = (rail_max.x-rail_min.x)*2;
			profile.append_vertex(Vector(rail_w*-0.5, 0, 0), false);
			profile.append_vertex(Vector(rail_w*0.5, 0, 0), false);
		}
		else
		{
			profile.append_vertex(Vector(i*(gauge*0.5+rail_min.x*2), 0, 0), false);
			profile.append_vertex(Vector(i*(gauge*0.5+rail_max.x*2), 0, 0), false);
		}

		// TODO Avoid generating redundant meshes for single-path tracks
		for(int j=-1; j<=4; ++j)
		{
			if(j>=0 && !((paths>>j)&1))
				continue;

			GL::Mesh *m = new GL::Mesh(GL::VERTEX3);
			GL::MeshBuilder bld(*m);
			unsigned index = 0;
			for(vector<TrackPart>::const_iterator k=parts.begin(); k!=parts.end(); ++k)
				if(j<0 || k->get_path()==static_cast<unsigned>(j))
					build_part(*k, profile, Vector(0, 0, ballast_h+1.5*rail_h), false, bld, index);
			path_meshes[(j&0xFF)|((i&3)<<8)] = m;
		}
	}
}

TrackType3D::~TrackType3D()
{
	for(map<unsigned, GL::Mesh *>::iterator i=path_meshes.begin(); i!=path_meshes.end(); ++i)
		delete i->second;
	if(own_data)
	{
		delete object;
		delete mesh;
	}
}

const GL::Mesh &TrackType3D::get_path_mesh(int p, int s) const
{
	unsigned key = (p<0 ? 0xFF : p) | ((s&3)<<8);
	return *get_item(path_meshes, key);
}

void TrackType3D::build_part(const TrackPart &part, const Profile &profile, const Vector &offset, bool mirror, GL::MeshBuilder &bld, unsigned &base_index)
{
	float plen = part.get_length();
	unsigned nsegs = (part.is_curved() ? static_cast<unsigned>(plen*32)+1 : 1);

	unsigned n_vertices = profile.get_n_vertices();
	for(unsigned i=0; i<=nsegs; ++i)
	{
		OrientedPoint basep = part.get_point(i*plen/nsegs);
		Transform trans = Transform::rotation(basep.rotation, Vector(0, 0, 1));

		for(unsigned j=0; j<n_vertices; ++j)
		{
			const Profile::Vertex &v = profile.get_vertex(mirror ? n_vertices-1-j : j);
			Vector p(0, -v.pos.x, v.pos.y);
			if(mirror)
				p.y = -p.y;
			p = basep.position+trans.transform(offset+p);

			Vector n(0, -v.normal.x, v.normal.y);
			if(mirror)
				n.y = -n.y;
			n = trans.transform(n);

			bld.normal(n.x, n.y, n.z);
			bld.vertex(p.x, p.y, p.z);
		}
	}

	for(unsigned i=0; i+1<n_vertices; )
	{
		bld.begin(GL::TRIANGLE_STRIP);
		for(unsigned j=0; j<=nsegs; ++j)
		{
			unsigned k = j*n_vertices+i;
			bld.element(base_index+k+1);
			bld.element(base_index+k);
		}
		bld.end();

		++i;
		if(!profile.get_vertex(i).smooth)
			++i;
	}

	base_index += (nsegs+1)*n_vertices;
}

void TrackType3D::build_ties(const TrackPart &part, const Profile &profile, float length, float spacing, const Vector &offset, GL::MeshBuilder &bld, unsigned &base_index)
{
	float plen = part.get_length();
	unsigned count = plen/spacing+0.5;

	unsigned n_vertices = profile.get_n_vertices();
	float min_y = profile.get_min_coords().y;
	vector<Vector> cap_vertices;
	for(unsigned i=0; i+1<n_vertices; )
	{
		const Profile::Vertex &v = profile.get_vertex(i);
		if(cap_vertices.size()>=2)
		{
			if(i+2<n_vertices)
				cap_vertices.push_back(Vector(v.pos.x, min_y, 0));
			else
				cap_vertices.push_back(profile.get_vertex(n_vertices-1).pos);
		}
		cap_vertices.push_back(v.pos);

		++i;
		if(!profile.get_vertex(i).smooth)
			++i;
	}

	for(unsigned i=0; i<count; ++i)
	{
		OrientedPoint basep = part.get_point((i+0.5)*plen/count);
		Transform trans = Transform::rotation(basep.rotation, Vector(0, 0, 1));

		unsigned strip = 0;
		for(unsigned j=0; j<n_vertices; ++j)
		{
			const Profile::Vertex &v = profile.get_vertex(j);

			Vector n = trans.transform(Vector(0, -v.normal.x, v.normal.y));
			bld.normal(n.x, n.y, n.z);

			if(!strip)
				bld.begin(GL::TRIANGLE_STRIP);

			for(unsigned k=0; k<2; ++k)
			{
				Vector p((k-0.5)*length, -v.pos.x, v.pos.y);
				p = basep.position+trans.transform(offset+p);
				bld.vertex(p.x, p.y, p.z);
			}

			if(++strip>=2 && !profile.get_vertex(j).smooth)
			{
				bld.end();
				strip = 0;
			}
		}
		if(strip)
			bld.end();

		for(unsigned j=0; j<2; ++j)
		{
			Vector n = trans.transform(Vector(j*2.0-1, 0, 0));
			bld.normal(n.x, n.y, n.z);

			bld.begin(GL::TRIANGLE_STRIP);
			for(unsigned k=0; k<cap_vertices.size(); ++k)
			{
				const Vector &v = cap_vertices[k^j];
				Vector p((j-0.5)*length, -v.x, v.y);
				p = basep.position+trans.transform(offset+p);
				bld.vertex(p.x, p.y, p.z);
			}
			bld.end();
		}

		base_index += 2*(n_vertices+cap_vertices.size());
	}
}

} // namespace R2C2
