#include <algorithm>
#include <msp/core/raii.h>
#include <msp/strings/format.h>
#include <msp/time/units.h>
#include "block.h"
#include "layout.h"
#include "route.h"
#include "trackcircuit.h"
#include "trackiter.h"
#include "tracktype.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

Block::Block(Layout &l, Track &start):
	TrackChain(l),
	id(0),
	sensor_addr(start.get_sensor_address()),
	turnout_addr(start.get_turnout_address()),
	conflict(false),
	sensor(0),
	train(0)
{
	add_track(start);

	if(start.get_type().is_turnout())
	{
		unsigned nls = start.get_n_link_slots();
		for(unsigned i=0; i<nls; ++i)
			endpoints.push_back(Endpoint(&start, i));
	}
	else
	{
		unsigned nls = start.get_n_link_slots();
		for(unsigned i=0; i<nls; ++i)
		{
			TrackIter iter = TrackIter(&start, i).flip();
			for(; (iter && check_validity(*iter)==VALID); iter=iter.next())
				add_track(*iter);
			if((iter = iter.flip()))
				endpoints.push_back(Endpoint(iter.track(), iter.entry()));
		}
	}

	determine_id();

	const set<Block *> &blocks = layout.get_all<Block>();
	for(set<Block *>::const_iterator i=blocks.begin(); (!conflict && i!=blocks.end()); ++i)
		conflict = (id==(*i)->get_id());

	if(!conflict && sensor_addr)
		sensor = new TrackCircuit(layout, *this);

	layout.add(*this);
}

Block::~Block()
{
	set<Track *> trks = tracks;
	tracks.clear();
	for(set<Track *>::iterator i=trks.begin(); i!=trks.end(); ++i)
		(*i)->set_block(0);

	for(vector<Endpoint>::iterator i=endpoints.begin(); i!=endpoints.end(); ++i)
		if(Block *blk = i->link)
		{
			i->link = 0;
			blk->break_link(*this);
		}

	layout.remove(*this);

	delete sensor;
}

void Block::set_name(const string &)
{
	throw logic_error("Block names can't be set");
}

void Block::on_track_added(Track &track)
{
	track.set_block(this);
}

TrackChain::Validity Block::check_validity(Track &track) const
{
	if(track.get_sensor_address()!=sensor_addr || track.get_turnout_address()!=turnout_addr)
		return INCOMPATIBLE;

	return TrackChain::check_validity(track);
}

const Block::Endpoint &Block::get_endpoint(unsigned i) const
{
	if(i>=endpoints.size())
		throw out_of_range("Block::get_endpoint");

	return endpoints[i];
}

int Block::get_endpoint_by_link(Block &other) const
{
	for(unsigned i=0; i<endpoints.size(); ++i)
		if(endpoints[i].link==&other)
			return i;

	return -1;
}

float Block::get_path_length(unsigned entry, const Route *route) const
{
	if(entry>=endpoints.size())
		throw out_of_range("Block::get_path_length");

	TrackIter t_iter = endpoints[entry].track_iter();

	float result = 0;
	while(t_iter && has_track(*t_iter))
	{
		unsigned path = (route ? route->get_path(*t_iter) : t_iter->get_active_path());
		result += t_iter->get_type().get_path_length(path);

		t_iter = t_iter.next(path);
	}

	return result;
}

void Block::check_link(Block &other)
{
	for(vector<Endpoint>::iterator i=endpoints.begin(); i!=endpoints.end(); ++i)
	{
		if(i->link)
			continue;

		for(vector<Endpoint>::iterator j=other.endpoints.begin(); j!=other.endpoints.end(); ++j)
			if(j->track==i->track->get_link(i->track_ep) && j->track->get_link(j->track_ep)==i->track && !j->link)
			{
				i->link = &other;
				j->link = this;

				determine_id();
				other.determine_id();
			}
	}
}

void Block::break_link(Block &other)
{
	for(vector<Endpoint>::iterator i=endpoints.begin(); i!=endpoints.end(); ++i)
		if(i->link==&other)
		{
			i->link = 0;
			other.break_link(*this);
			determine_id();
		}
}

Block *Block::get_link(unsigned epi) const
{
	if(epi>=endpoints.size())
		throw out_of_range("Block::get_link");
	return endpoints[epi].link;
}

bool Block::reserve(Train *t)
{
	if(!t || !train)
	{
		train = t;
		signal_reserved.emit(t);
		return true;
	}
	else
		return false;
}

void Block::determine_id()
{
	string n;
	if(sensor_addr)
	{
		id = 0x1000|sensor_addr;
		n = format("Sensor %d", sensor_addr);
	}
	else if(turnout_addr)
	{
		id = 0x2000|turnout_addr;
		n = format("Turnout %d", turnout_addr);
	}
	else if(endpoints.size()==2)
	{
		unsigned id1 = endpoints[0].link ? endpoints[0].link->get_id() : 1;
		unsigned id2 = endpoints[1].link ? endpoints[1].link->get_id() : 1;
		if(id2<id1)
			swap(id1, id2);
		id = (id1<<16)|id2;
	}
	else if(endpoints.size()==1)
	{
		unsigned id1 = endpoints[0].link ? endpoints[0].link->get_id() : 1;
		id = 0x10000 | id1;
	}

	if(n.empty())
		n = format("Block %x", id);

	name = n;
	signal_name_changed.emit(name);
}

DataFile::Statement Block::save_reference() const
{
	return (DataFile::Statement("block"), id);
}


Block::Endpoint::Endpoint(Track *t, unsigned e):
	track(t),
	track_ep(e),
	link(0)
{ }

TrackIter Block::Endpoint::track_iter() const
{
	return TrackIter(track, track_ep);
}

} // namespace R2C2
